@@ -76,6 +76,8 @@ char upgrade_from_version[sizeof("10.20.456-MariaDB")+30];
 
 static my_bool opt_write_binlog;
 
+static void print_conn_args(const char *tool_name);
+
 #define OPT_SILENT OPT_MAX_CLIENT_OPTION
 
 static struct my_option my_long_options[]=
@@ -155,7 +157,10 @@ static struct my_option my_long_options[]=
    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
   {"user", 'u', "User for login.", &opt_user,
    &opt_user, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
-  {"verbose", 'v', "Display more output about the process; Using it twice will print connection argument; Using it 3 times will print out all CHECK, RENAME and ALTER TABLE during the check phase.",
+  {"verbose", 'v', "Display more output about the process; Using it twice will print connection argument;"
+  "Using it 3 times will print out all CHECK, RENAME and ALTER TABLE during the check phase;"
+  "Using it 4 times (added in MariaDB 10.0.14) will also write out all mariadb-check commands used;"
+  "Using it 5 times will print all the mariadb commands used and their results while running mysql_fix_privilege_tables script.",
    &opt_not_used, &opt_not_used, 0, GET_BOOL, NO_ARG, 1, 0, 0, 0, 0, 0},
   {"version", 'V', "Output version information and exit.", 0, 0, 0,
    GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0},
@@ -210,6 +215,7 @@ static void die(const char *fmt, ...)
   DBUG_ENTER("die");
 
   /* Print the error message */
+  print_conn_args("mariadb-check");
   fflush(stdout);
   va_start(args, fmt);
   if (fmt)
@@ -640,6 +646,7 @@ static int run_query(const char *query, DYNAMIC_STRING *ds_res,
                 "--database=mysql",
                 "--batch", /* Turns off pager etc. */
                 force ? "--force": "--skip-force",
+                opt_verbose >= 5 ? "--verbose" : "",
                 ds_res || opt_silent ? "--silent": "",
                 "<",
                 query_file_path,
@@ -866,8 +873,7 @@ static int upgrade_already_done(int silent)
             "There is no need to run mysql_upgrade again for %s.",
             upgrade_from_version, version);
     if (!opt_check_upgrade)
-      verbose("You can use --force if you still want to run mysql_upgrade",
-              upgrade_from_version, version);
+      verbose("You can use --force if you still want to run mysql_upgrade");
   }
   return 0;
 }
@@ -935,7 +941,7 @@ static int run_mysqlcheck_upgrade(my_bool mysql_db_only)
     return 0;
   }
   verbose("Phase %d/%d: Checking and upgrading %s", ++phase, phases_total, what);
-  print_conn_args("mysqlcheck");
+  print_conn_args("mariadb-check");
   retch= run_tool(mysqlcheck_path,
                   NULL, /* Send output from mysqlcheck directly to screen */
                   defaults_file,
@@ -1149,6 +1155,8 @@ static int install_used_plugin_data_types(void)
   DYNAMIC_STRING ds_result;
   const char *query = "SELECT table_comment FROM information_schema.tables"
                       " WHERE table_comment LIKE 'Unknown data type: %'";
+  if (opt_systables_only)
+    return 0;
   if (init_dynamic_string(&ds_result, "", 512, 512))
     die("Out of memory");
   run_query(query, &ds_result, TRUE);
@@ -1329,9 +1337,7 @@ static int run_sql_fix_privilege_tables(void)
       dynstr_append(&ds_script, *query_ptr);
   }
 
-  run_query(ds_script.str,
-            &ds_result, /* Collect result */
-            TRUE);
+  run_query(ds_script.str, (opt_verbose >= 5) ? NULL : &ds_result, TRUE);
 
   {
     /*
@@ -1432,7 +1438,7 @@ int main(int argc, char **argv)
 
   if (tty_password)
   {
-    opt_password= get_tty_password(NullS);
+    opt_password= my_get_tty_password(NullS);
     /* add password to defaults file */
     add_one_option_cnf_file(&ds_args, "password", opt_password);
   }
@@ -1451,15 +1457,20 @@ int main(int argc, char **argv)
   }
 
   /* Find mysql */
-  find_tool(mysql_path, IF_WIN("mysql.exe", "mysql"), self_name);
+  find_tool(mysql_path, IF_WIN("mariadb.exe", "mariadb"), self_name);
 
   open_mysql_upgrade_file();
 
   if (opt_check_upgrade)
-    exit(upgrade_already_done(0) == 0);
+  {
+    int upgrade_needed = upgrade_already_done(0);
+    free_used_memory();
+    my_end(my_end_arg);
+    exit(upgrade_needed == 0);
+  }
 
   /* Find mysqlcheck */
-  find_tool(mysqlcheck_path, IF_WIN("mysqlcheck.exe", "mysqlcheck"), self_name);
+  find_tool(mysqlcheck_path, IF_WIN("mariadb-check.exe", "mariadb-check"), self_name);
 
   if (opt_systables_only && !opt_silent)
     printf("The --upgrade-system-tables option was used, user tables won't be touched.\n");
@@ -1502,6 +1513,7 @@ int main(int argc, char **argv)
   DBUG_ASSERT(phase == phases_total);
 
 end:
+  print_conn_args("mariadb-check");
   free_used_memory();
   my_end(my_end_arg);
   exit(0);
