@@ -824,8 +824,8 @@ bool mysql_derived_prepare(THD *thd, LEX *lex, TABLE_LIST *derived)
   if ((res= unit->prepare(derived, derived->derived_result, 0)))
     goto exit;
   if (derived->with &&
-      (res= derived->with->rename_columns_of_derived_unit(thd, unit)))
-    goto exit; 
+      (res= derived->with->process_columns_of_derived_unit(thd, unit)))
+    goto exit;
   if ((res= check_duplicate_names(thd, unit->types, 0)))
     goto exit;
 
@@ -1227,6 +1227,9 @@ bool mysql_derived_fill(THD *thd, LEX *lex, TABLE_LIST *derived)
       goto err;
     JOIN *join= unit->first_select()->join;
     join->first_record= false;
+    if (join->zero_result_cause)
+      goto err;
+
     for (uint i= join->top_join_tab_count;
          i < join->top_join_tab_count + join->aggr_tables;
          i++)
@@ -1258,13 +1261,12 @@ bool mysql_derived_fill(THD *thd, LEX *lex, TABLE_LIST *derived)
   {
     SELECT_LEX *first_select= unit->first_select();
     unit->set_limit(unit->global_parameters());
-    if (unit->select_limit_cnt == HA_POS_ERROR)
+    if (unit->lim.is_unlimited())
       first_select->options&= ~OPTION_FOUND_ROWS;
 
     lex->current_select= first_select;
     res= mysql_select(thd,
                       first_select->table_list.first,
-                      first_select->with_wild,
                       first_select->item_list, first_select->where,
                       (first_select->order_list.elements+
                        first_select->group_list.elements),
@@ -1336,6 +1338,10 @@ bool mysql_derived_reinit(THD *thd, LEX *lex, TABLE_LIST *derived)
                        derived->get_unit()));
   st_select_lex_unit *unit= derived->get_unit();
 
+  // reset item names to that saved after wildcard expansion in JOIN::prepare
+  for(st_select_lex *sl= unit->first_select(); sl; sl= sl->next_select())
+    sl->restore_item_list_names();
+
   derived->merged_for_insert= FALSE;
   unit->unclean();
   unit->types.empty();
