@@ -41,16 +41,14 @@ Created 3/26/1996 Heikki Tuuri
 #ifdef WITH_WSREP
 #include "trx0xa.h"
 #endif /* WITH_WSREP */
-
-typedef UT_LIST_BASE_NODE_T(trx_t) trx_ut_list_t;
+#include "ilist.h"
 
 /** Checks if a page address is the trx sys header page.
 @param[in]	page_id	page id
 @return true if trx sys header page */
-inline bool trx_sys_hdr_page(const page_id_t& page_id)
+inline bool trx_sys_hdr_page(const page_id_t page_id)
 {
-	return(page_id.space() == TRX_SYS_SPACE
-	       && page_id.page_no() == TRX_SYS_PAGE_NO);
+  return page_id == page_id_t(TRX_SYS_SPACE, TRX_SYS_PAGE_NO);
 }
 
 /*****************************************************************//**
@@ -340,9 +338,6 @@ FIL_PAGE_ARCH_LOG_NO_OR_SPACE_ID. */
 constexpr uint32_t TRX_SYS_DOUBLEWRITE_MAGIC_N= 536853855;
 /** Contents of TRX_SYS_DOUBLEWRITE_SPACE_ID_STORED */
 constexpr uint32_t TRX_SYS_DOUBLEWRITE_SPACE_ID_STORED_N= 1783657386;
-
-/** Size of the doublewrite block in pages */
-#define TRX_SYS_DOUBLEWRITE_BLOCK_SIZE	FSP_EXTENT_SIZE
 /* @} */
 
 trx_t* current_trx();
@@ -362,6 +357,13 @@ struct rw_trx_hash_element_t
 
 
   trx_id_t id; /* lf_hash_init() relies on this to be first in the struct */
+
+  /**
+    Transaction serialization number.
+
+    Assigned shortly before the transaction is moved to COMMITTED_IN_MEMORY
+    state. Initially set to TRX_ID_MAX.
+  */
   Atomic_counter<trx_id_t> no;
   trx_t *trx;
   ib_mutex_t mutex;
@@ -377,6 +379,10 @@ class rw_trx_hash_t
   LF_HASH hash;
 
 
+  template <typename T>
+  using walk_action= my_bool(rw_trx_hash_element_t *element, T *action);
+
+
   /**
     Constructor callback for lock-free allocator.
 
@@ -479,18 +485,19 @@ class rw_trx_hash_t
   }
 
 
-  struct eliminate_duplicates_arg
+  template <typename T> struct eliminate_duplicates_arg
   {
     trx_ids_t ids;
-    my_hash_walk_action action;
-    void *argument;
-    eliminate_duplicates_arg(size_t size, my_hash_walk_action act, void* arg):
+    walk_action<T> *action;
+    T *argument;
+    eliminate_duplicates_arg(size_t size, walk_action<T> *act, T *arg):
       action(act), argument(arg) { ids.reserve(size); }
   };
 
 
+  template <typename T>
   static my_bool eliminate_duplicates(rw_trx_hash_element_t *element,
-                                      eliminate_duplicates_arg *arg)
+                                      eliminate_duplicates_arg<T> *arg)
   {
     for (trx_ids_t::iterator it= arg->ids.begin(); it != arg->ids.end(); it++)
     {
@@ -517,15 +524,16 @@ class rw_trx_hash_t
   }
 
 
-  struct debug_iterator_arg
+  template <typename T> struct debug_iterator_arg
   {
-    my_hash_walk_action action;
-    void *argument;
+    walk_action<T> *action;
+    T *argument;
   };
 
 
+  template <typename T>
   static my_bool debug_iterator(rw_trx_hash_element_t *element,
-                                debug_iterator_arg *arg)
+                                debug_iterator_arg<T> *arg)
   {
     mutex_enter(&element->mutex);
     if (element->trx)
@@ -759,23 +767,28 @@ public:
       @retval 1 iteration was interrupted (action returned 1)
   */
 
-  int iterate(trx_t *caller_trx, my_hash_walk_action action, void *argument)
+  template <typename T>
+  int iterate(trx_t *caller_trx, walk_action<T> *action, T *argument= nullptr)
   {
     LF_PINS *pins= caller_trx ? get_pins(caller_trx) : lf_hash_get_pins(&hash);
     ut_a(pins);
 #ifdef UNIV_DEBUG
-    debug_iterator_arg debug_arg= { action, argument };
-    action= reinterpret_cast<my_hash_walk_action>(debug_iterator);
-    argument= &debug_arg;
+    debug_iterator_arg<T> debug_arg= { action, argument };
+    action= reinterpret_cast<decltype(action)>(debug_iterator<T>);
+    argument= reinterpret_cast<T*>(&debug_arg);
 #endif
-    int res= lf_hash_iterate(&hash, pins, action, argument);
+    int res= lf_hash_iterate(&hash, pins,
+                             reinterpret_cast<my_hash_walk_action>(action),
+                             const_cast<void*>(static_cast<const void*>
+                             (argument)));
     if (!caller_trx)
       lf_hash_put_pins(pins);
     return res;
   }
 
 
-  int iterate(my_hash_walk_action action, void *argument)
+  template <typename T>
+  int iterate(walk_action<T> *action, T *argument= nullptr)
   {
     return iterate(current_trx(), action, argument);
   }
@@ -787,21 +800,73 @@ public:
     @sa iterate()
   */
 
-  int iterate_no_dups(trx_t *caller_trx, my_hash_walk_action action,
-                      void *argument)
+  template <typename T>
+  int iterate_no_dups(trx_t *caller_trx, walk_action<T> *action,
+                      T *argument= nullptr)
   {
-    eliminate_duplicates_arg arg(size() + 32, action, argument);
-    return iterate(caller_trx, reinterpret_cast<my_hash_walk_action>
-                   (eliminate_duplicates), &arg);
+    eliminate_duplicates_arg<T> arg(size() + 32, action, argument);
+    return iterate(caller_trx, eliminate_duplicates<T>, &arg);
   }
 
 
-  int iterate_no_dups(my_hash_walk_action action, void *argument)
+  template <typename T>
+  int iterate_no_dups(walk_action<T> *action, T *argument= nullptr)
   {
     return iterate_no_dups(current_trx(), action, argument);
   }
 };
 
+class thread_safe_trx_ilist_t
+{
+public:
+  void create() { mutex_create(LATCH_ID_TRX_SYS, &mutex); }
+  void close() { mutex_free(&mutex); }
+
+  bool empty() const
+  {
+    mutex_enter(&mutex);
+    auto result= trx_list.empty();
+    mutex_exit(&mutex);
+    return result;
+  }
+
+  void push_front(trx_t &trx)
+  {
+    mutex_enter(&mutex);
+    trx_list.push_front(trx);
+    mutex_exit(&mutex);
+  }
+
+  void remove(trx_t &trx)
+  {
+    mutex_enter(&mutex);
+    trx_list.remove(trx);
+    mutex_exit(&mutex);
+  }
+
+  template <typename Callable> void for_each(Callable &&callback) const
+  {
+    mutex_enter(&mutex);
+    for (const auto &trx : trx_list)
+      callback(trx);
+    mutex_exit(&mutex);
+  }
+
+  template <typename Callable> void for_each(Callable &&callback)
+  {
+    mutex_enter(&mutex);
+    for (auto &trx : trx_list)
+      callback(trx);
+    mutex_exit(&mutex);
+  }
+
+  void freeze() const { mutex_enter(&mutex); }
+  void unfreeze() const { mutex_exit(&mutex); }
+
+private:
+  alignas(CACHE_LINE_SIZE) mutable TrxSysMutex mutex;
+  alignas(CACHE_LINE_SIZE) ilist<trx_t> trx_list;
+};
 
 /** The transaction system central memory data structure. */
 class trx_sys_t
@@ -832,11 +897,8 @@ public:
   */
   MY_ALIGNED(CACHE_LINE_SIZE) Atomic_counter<size_t> rseg_history_len;
 
-  /** Mutex protecting trx_list. */
-  MY_ALIGNED(CACHE_LINE_SIZE) mutable TrxSysMutex mutex;
-
   /** List of all transactions. */
-  MY_ALIGNED(CACHE_LINE_SIZE) trx_ut_list_t trx_list;
+  thread_safe_trx_ilist_t trx_list;
 
 	MY_ALIGNED(CACHE_LINE_SIZE)
 	/** Temporary rollback segments */
@@ -894,8 +956,7 @@ public:
   trx_id_t get_min_trx_id()
   {
     trx_id_t id= get_max_trx_id();
-    rw_trx_hash.iterate(reinterpret_cast<my_hash_walk_action>
-                        (get_min_trx_id_callback), &id);
+    rw_trx_hash.iterate(get_min_trx_id_callback, &id);
     return id;
   }
 
@@ -950,8 +1011,7 @@ public:
   */
   void assign_new_trx_no(trx_t *trx)
   {
-    trx->no= get_new_trx_id_no_refresh();
-    trx->rw_trx_hash_element->no= trx->no;
+    trx->rw_trx_hash_element->no= get_new_trx_id_no_refresh();
     refresh_rw_trx_hash_version();
   }
 
@@ -975,13 +1035,12 @@ public:
     @param[in,out] caller_trx used to get access to rw_trx_hash_pins
     @param[out]    ids        array to store registered transaction identifiers
     @param[out]    max_trx_id variable to store m_max_trx_id value
-    @param[out]    mix_trx_no variable to store min(trx->no) value
+    @param[out]    mix_trx_no variable to store min(no) value
   */
 
   void snapshot_ids(trx_t *caller_trx, trx_ids_t *ids, trx_id_t *max_trx_id,
                     trx_id_t *min_trx_no)
   {
-    ut_ad(!mutex_own(&mutex));
     snapshot_ids_arg arg(ids);
 
     while ((arg.m_id= get_rw_trx_hash_version()) != get_max_trx_id())
@@ -990,9 +1049,7 @@ public:
 
     ids->clear();
     ids->reserve(rw_trx_hash.size() + 32);
-    rw_trx_hash.iterate(caller_trx,
-                        reinterpret_cast<my_hash_walk_action>(copy_one_id),
-                        &arg);
+    rw_trx_hash.iterate(caller_trx, copy_one_id, &arg);
 
     *max_trx_id= arg.m_id;
     *min_trx_no= arg.m_no;
@@ -1017,7 +1074,7 @@ public:
   void close();
 
   /** @return total number of active (non-prepared) transactions */
-  ulint any_active_transactions();
+  size_t any_active_transactions(size_t *prepared= nullptr);
 
 
   /**
@@ -1080,9 +1137,7 @@ public:
   */
   void register_trx(trx_t *trx)
   {
-    mutex_enter(&mutex);
-    UT_LIST_ADD_FIRST(trx_list, trx);
-    mutex_exit(&mutex);
+    trx_list.push_front(*trx);
   }
 
 
@@ -1093,9 +1148,7 @@ public:
   */
   void deregister_trx(trx_t *trx)
   {
-    mutex_enter(&mutex);
-    UT_LIST_REMOVE(trx_list, trx);
-    mutex_exit(&mutex);
+    trx_list.remove(*trx);
   }
 
 
@@ -1106,7 +1159,7 @@ public:
     in. This function is called by purge thread to determine whether it should
     purge the delete marked record or not.
   */
-  void clone_oldest_view();
+  void clone_oldest_view(ReadViewBase *view) const;
 
 
   /** @return the number of active views */
@@ -1114,14 +1167,11 @@ public:
   {
     size_t count= 0;
 
-    mutex_enter(&mutex);
-    for (const trx_t *trx= UT_LIST_GET_FIRST(trx_list); trx;
-         trx= UT_LIST_GET_NEXT(trx_list, trx))
-    {
-      if (trx->read_view.get_state() == READ_VIEW_STATE_OPEN)
+    trx_list.for_each([&count](const trx_t &trx) {
+      if (trx.read_view.is_open())
         ++count;
-    }
-    mutex_exit(&mutex);
+    });
+
     return count;
   }
 
