@@ -10,9 +10,14 @@
 #include <my_global.h>
 #include "mysql_version.h"
 #include "hs_compat.h"
+#if MYSQL_VERSION_ID < 50500
+#include "mysql_priv.h"
+#include <mysql/plugin.h>
+#else
 #include "sql_priv.h"
 #include "probes_mysql.h"
 #include "sql_class.h"
+#endif
 
 #include "hstcpcli.hpp"
 #include "auto_file.hpp"
