@@ -1,4 +1,5 @@
-/* Copyright (C) 2008-2018 Kentoku Shiba
+/* Copyright (C) 2008-2020 Kentoku Shiba
+   Copyright (C) 2019, 2022, MariaDB Corporation.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -20,34 +21,63 @@
 
 #define SPIDER_DBTON_SIZE 15
 
+#ifndef SIZEOF_STORED_DOUBLE
+#define SIZEOF_STORED_DOUBLE 8
+#endif
+
 #define SPIDER_DB_WRAPPER_MYSQL "mysql"
 #define SPIDER_DB_WRAPPER_MARIADB "mariadb"
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100204
 #define PLUGIN_VAR_CAN_MEMALLOC
 /*
+#define ITEM_FUNC_CASE_PARAMS_ARE_PUBLIC
 #define HASH_UPDATE_WITH_HASH_VALUE
 */
+#else
+#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+#define HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
+#endif
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100002
 #define SPIDER_HAS_DISCOVER_TABLE_STRUCTURE
 #define SPIDER_HAS_APPEND_FOR_SINGLE_QUOTE
 #define SPIDER_HAS_SHOW_SIMPLE_FUNC
 #define SPIDER_HAS_JT_HASH_INDEX_MERGE
 #define SPIDER_HAS_EXPR_CACHE_ITEM
+#else
+#define SPIDER_NEED_CHECK_CONDITION_AT_CHECKING_DIRECT_ORDER_LIMIT
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100007
 #define SPIDER_ITEM_HAS_CMP_TYPE
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100004
 #define SPIDER_HAS_TIME_STATUS
 #define SPIDER_HAS_DECIMAL_OPERATION_RESULTS_VALUE_TYPE
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100014
 #define SPIDER_ITEM_STRING_WITHOUT_SET_STR_WITH_COPY
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100100
 #define SPIDER_ITEM_STRING_WITHOUT_SET_STR_WITH_COPY_AND_THDPTR
+#endif
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100108
 #define SPIDER_HAS_GROUP_BY_HANDLER
+#endif
 
+#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100200
 #define SPIDER_ORDER_HAS_ENUM_ORDER
+#endif
 
+#if defined(MARIADB_BASE_VERSION)
 #define SPIDER_ITEM_GEOFUNC_NAME_HAS_MBR
 #define SPIDER_HANDLER_AUTO_REPAIR_HAS_ERROR
+#endif
 
 class spider_db_conn;
 typedef spider_db_conn SPIDER_DB_CONN;
@@ -64,6 +94,8 @@ typedef st_spider_result SPIDER_RESULT;
 
 #define SPIDER_SQL_SEMICOLON_STR ";"
 #define SPIDER_SQL_SEMICOLON_LEN sizeof(SPIDER_SQL_SEMICOLON_STR) - 1
+#define SPIDER_SQL_COLON_STR ":"
+#define SPIDER_SQL_COLON_LEN sizeof(SPIDER_SQL_COLON_STR) - 1
 #define SPIDER_SQL_VALUE_QUOTE_STR "'"
 #define SPIDER_SQL_VALUE_QUOTE_LEN (sizeof(SPIDER_SQL_VALUE_QUOTE_STR) - 1)
 
@@ -71,6 +103,8 @@ typedef st_spider_result SPIDER_RESULT;
 #define SPIDER_SQL_DOT_LEN (sizeof(SPIDER_SQL_DOT_STR) - 1)
 #define SPIDER_SQL_PERCENT_STR "%"
 #define SPIDER_SQL_PERCENT_LEN (sizeof(SPIDER_SQL_PERCENT_STR) - 1)
+#define SPIDER_SQL_HYPHEN_STR "-"
+#define SPIDER_SQL_HYPHEN_LEN (sizeof(SPIDER_SQL_HYPHEN_STR) - 1)
 
 #define SPIDER_SQL_EQUAL_STR " = "
 #define SPIDER_SQL_EQUAL_LEN (sizeof(SPIDER_SQL_EQUAL_STR) - 1)
@@ -105,6 +139,7 @@ typedef st_spider_result SPIDER_RESULT;
 #define SPIDER_SQL_HS_LTEQUAL_STR "<="
 #define SPIDER_SQL_HS_LTEQUAL_LEN (sizeof(SPIDER_SQL_HS_LTEQUAL_STR) - 1)
 
+#ifdef ITEM_FUNC_CASE_PARAMS_ARE_PUBLIC
 #define SPIDER_SQL_CASE_STR "case "
 #define SPIDER_SQL_CASE_LEN (sizeof(SPIDER_SQL_CASE_STR) - 1)
 #define SPIDER_SQL_WHEN_STR " when "
@@ -115,6 +150,7 @@ typedef st_spider_result SPIDER_RESULT;
 #define SPIDER_SQL_ELSE_LEN (sizeof(SPIDER_SQL_ELSE_STR) - 1)
 #define SPIDER_SQL_END_STR " end"
 #define SPIDER_SQL_END_LEN (sizeof(SPIDER_SQL_END_STR) - 1)
+#endif
 
 #define SPIDER_SQL_USING_STR " using "
 #define SPIDER_SQL_USING_LEN (sizeof(SPIDER_SQL_USING_STR) - 1)
@@ -186,6 +222,11 @@ typedef st_spider_result SPIDER_RESULT;
 #define SPIDER_SQL_CONNECTION_LEN (sizeof(SPIDER_SQL_CONNECTION_STR) - 1)
 #define SPIDER_SQL_LCL_NAME_QUOTE_STR "`"
 #define SPIDER_SQL_LCL_NAME_QUOTE_LEN (sizeof(SPIDER_SQL_LCL_NAME_QUOTE_STR) - 1)
+#define SPIDER_SQL_MIN_STR "min"
+#define SPIDER_SQL_MIN_LEN (sizeof(SPIDER_SQL_MIN_STR) - 1)
+
+#define SPIDER_SQL_LOP_CHK_PRM_PRF_STR "spider_lc_"
+#define SPIDER_SQL_LOP_CHK_PRM_PRF_LEN (sizeof(SPIDER_SQL_LOP_CHK_PRM_PRF_STR) - 1)
 
 #define SPIDER_CONN_KIND_MYSQL (1 << 0)
 #if defined(HS_HAS_SQLCOM) && defined(HAVE_HANDLERSOCKET)
@@ -811,6 +852,10 @@ public:
     spider_string *str,
     Time_zone *time_zone
   ) = 0;
+  virtual int append_loop_check(
+    spider_string *str,
+    SPIDER_CONN *conn
+  );
   virtual int append_start_transaction(
     spider_string *str
   ) = 0;
@@ -876,6 +921,11 @@ public:
     spider_string *str
   ) = 0;
 #endif
+  virtual bool tables_on_different_db_are_joinable();
+  virtual bool socket_has_default_value();
+  virtual bool database_has_default_value();
+  virtual bool append_charset_name_before_string();
+  virtual uint limit_mode();
 };
 
 class spider_db_row
@@ -926,12 +976,12 @@ public:
 
 class spider_db_result
 {
-protected:
-  SPIDER_DB_CONN *db_conn;
 public:
-  uint dbton_id;
+  SPIDER_DB_CONN *db_conn;
+  uint           dbton_id;
   spider_db_result(SPIDER_DB_CONN *in_db_conn);
   virtual ~spider_db_result() = default;
+  virtual void set_limit(longlong value) {}
   virtual bool has_result() = 0;
   virtual void free_result() = 0;
   virtual SPIDER_DB_ROW *current_row() = 0;
@@ -986,19 +1036,20 @@ public:
     CHARSET_INFO *access_charset
   ) = 0;
 #endif
+  virtual uint limit_mode();
 };
 
 class spider_db_conn
 {
-protected:
-  SPIDER_CONN    *conn;
 public:
-  uint dbton_id;
+  SPIDER_CONN    *conn;
+  uint           dbton_id;
   spider_db_conn(
     SPIDER_CONN *in_conn
   );
   virtual ~spider_db_conn() = default;
   virtual int init() = 0;
+  virtual void set_limit(longlong value) {}
   virtual bool is_connected() = 0;
   virtual void bg_connect() = 0;
   virtual int connect(
@@ -1037,6 +1088,7 @@ public:
     int *error_num
   ) = 0;
   virtual spider_db_result *use_result(
+    ha_spider *spider,
     st_spider_db_request_key *request_key,
     int *error_num
   ) = 0;
@@ -1118,6 +1170,11 @@ public:
     Time_zone *time_zone,
     int *need_mon
   ) = 0;
+  virtual bool set_loop_check_in_bulk_sql();
+  virtual int set_loop_check(
+    int *need_mon
+  );
+  virtual int fin_loop_check();
   virtual int show_master_status(
     SPIDER_TRX *trx,
     SPIDER_SHARE *share,
@@ -1200,6 +1257,7 @@ public:
   virtual bool cmp_request_key_to_snd(
     st_spider_db_request_key *request_key
   ) = 0;
+  virtual uint limit_mode();
 };
 
 class spider_db_share
@@ -1259,6 +1317,7 @@ public:
 #ifdef SPIDER_HAS_GROUP_BY_HANDLER
   SPIDER_LINK_IDX_CHAIN *link_idx_chain;
 #endif
+  bool strict_group_by;
   bool no_where_cond;
   spider_db_handler(ha_spider *spider, spider_db_share *db_share) :
     dbton_id(db_share->dbton_id), spider(spider), db_share(db_share),
@@ -1503,8 +1562,8 @@ public:
     ulong sql_type
   ) = 0;
   virtual int append_explain_select_part(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     ulong sql_type,
     int link_idx
   ) = 0;
@@ -1618,8 +1677,8 @@ public:
     ulonglong &last_insert_id
   ) = 0;
   virtual ha_rows explain_select(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     int link_idx
   ) = 0;
   virtual int lock_tables(
@@ -1734,6 +1793,18 @@ public:
     ulong sql_type
   ) = 0;
 #endif
+#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+  virtual bool check_direct_update(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+  virtual bool check_direct_delete(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+#endif
 };
 
 class spider_db_copy_table
@@ -1809,6 +1880,23 @@ enum spider_db_access_type
   SPIDER_DB_ACCESS_TYPE_NOSQL
 };
 
+#define SPIDER_MATURITY_UNKNOWN      0
+#define SPIDER_MATURITY_EXPERIMENTAL 1
+#define SPIDER_MATURITY_ALPHA        2
+#define SPIDER_MATURITY_BETA         3
+#define SPIDER_MATURITY_GAMMA        4
+#define SPIDER_MATURITY_STABLE       5
+
+static const LEX_CSTRING maturity_name[] =
+{
+  { STRING_WITH_LEN("Unknown") },
+  { STRING_WITH_LEN("Experimental") },
+  { STRING_WITH_LEN("Alpha") },
+  { STRING_WITH_LEN("Beta") },
+  { STRING_WITH_LEN("Gamma") },
+  { STRING_WITH_LEN("Stable") }
+};
+
 /*
   Type of singletons based on the type of the remote database.
 
@@ -1830,6 +1918,9 @@ typedef struct st_spider_dbton
   SPIDER_DB_CONN *(*create_db_conn)(SPIDER_CONN *conn);
   bool (*support_direct_join)();
   spider_db_util *db_util;
+  const char *descr;
+  const char *version_info;
+  unsigned int maturity;
 } SPIDER_DBTON;
 
 typedef struct st_spider_position
@@ -1977,7 +2068,6 @@ typedef struct st_spider_result_list
 #endif
     int                   quick_phase;
   bool                    keyread;
-  int                     lock_type;
   TABLE                   *table;
 #ifndef WITHOUT_SPIDER_BG_SEARCH
   volatile int            bgs_error;
