@@ -87,6 +87,10 @@ public:
     spider_string *str,
     Time_zone *time_zone
   ) override;
+  int append_loop_check(
+    spider_string *str,
+    SPIDER_CONN *conn
+  ) override;
   int append_start_transaction(
     spider_string *str
   ) override;
@@ -187,6 +191,7 @@ public:
     spider_string *str
   ) override;
 #endif
+  bool append_charset_name_before_string() override;
 };
 
 class spider_db_mysql_util: public spider_db_mbase_util
@@ -433,13 +438,14 @@ public:
   bool is_xa_nota_error(
     int error_num
   );
-  void fetch_and_print_warnings(struct tm *l_time);
+  int fetch_and_print_warnings(struct tm *l_time);
   spider_db_result *store_result(
     spider_db_result_buffer **spider_res_buf,
     st_spider_db_request_key *request_key,
     int *error_num
   );
   spider_db_result *use_result(
+    ha_spider *spider,
     st_spider_db_request_key *request_key,
     int *error_num
   );
@@ -521,6 +527,11 @@ public:
     Time_zone *time_zone,
     int *need_mon
   );
+  bool set_loop_check_in_bulk_sql();
+  int set_loop_check(
+    int *need_mon
+  );
+  int fin_loop_check();
   int exec_simple_sql_with_result(
     SPIDER_TRX *trx,
     SPIDER_SHARE *share,
@@ -1376,15 +1387,15 @@ public:
     int link_idx
   );
   int append_explain_select_part(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     ulong sql_type,
     int link_idx
   );
   int append_explain_select(
     spider_string *str,
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     ulong sql_type,
     int link_idx
   );
@@ -1510,8 +1521,8 @@ public:
     ulonglong &last_insert_id
   );
   ha_rows explain_select(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     int link_idx
   );
   int lock_tables(
@@ -1651,6 +1662,18 @@ public:
     spider_fields *fields
   );
 #endif
+#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+  bool check_direct_update(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+  bool check_direct_delete(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+#endif
 };
 
 class spider_mysql_handler: public spider_mbase_handler
