@@ -1,5 +1,5 @@
-/* Copyright (C) 2010-2019 Kentoku Shiba
-   Copyright (C) 2019 MariaDB corp
+/* Copyright (C) 2010-2020 Kentoku Shiba
+   Copyright (C) 2019-2020 MariaDB corp
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -54,6 +54,7 @@ static LEX_STRING spider_init_queries[] = {
     "  ssl_verify_server_cert tinyint not null default 0,"
     "  default_file text,"
     "  default_group char(64) default null,"
+    "  dsn char(64) default null,"
     "  key idx1 (data, format_id, gtrid_length, host)"
     ") engine=MyISAM default charset=utf8 collate=utf8_bin"
   )},
@@ -77,6 +78,7 @@ static LEX_STRING spider_init_queries[] = {
     "  ssl_verify_server_cert tinyint not null default 0,"
     "  default_file text,"
     "  default_group char(64) default null,"
+    "  dsn char(64) default null,"
     "  thread_id int default null,"
     "  status char(8) not null default '',"
     "  failed_time timestamp not null default current_timestamp,"
@@ -105,6 +107,7 @@ static LEX_STRING spider_init_queries[] = {
     "  monitoring_binlog_pos_at_failing tinyint not null default 0,"
     "  default_file text,"
     "  default_group char(64) default null,"
+    "  dsn char(64) default null,"
     "  tgt_db_name char(64) default null,"
     "  tgt_table_name char(64) default null,"
     "  link_status tinyint not null default 1,"
@@ -136,6 +139,7 @@ static LEX_STRING spider_init_queries[] = {
     "  ssl_verify_server_cert tinyint not null default 0,"
     "  default_file text,"
     "  default_group char(64) default null,"
+    "  dsn char(64) default null,"
     "  primary key (db_name, table_name, link_id, sid)"
     ") engine=MyISAM default charset=utf8 collate=utf8_bin"
   )},
@@ -621,6 +625,25 @@ static LEX_STRING spider_init_queries[] = {
     "    algorithm=copy, lock=shared;"
     "end if;"
   )},
+/*
+  Fix for version 3.4
+*/
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_link_mon_servers"
+    "  add column if not exists dsn char(64) default null after default_group;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_tables"
+    "  add column if not exists dsn char(64) default null after default_group;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_xa_failed_log"
+    "  add column if not exists dsn char(64) default null after default_group;"
+  )},
+  {C_STRING_WITH_LEN(
+    "alter table mysql.spider_xa_member"
+    "  add column if not exists dsn char(64) default null after default_group;"
+  )},
   {C_STRING_WITH_LEN(
     "set @win_plugin := IF(@@version_compile_os like 'Win%', 1, 0);"
   )},
