@@ -56,8 +56,10 @@
 #include "spd_trx.h"
 
 extern struct st_mysql_plugin spider_i_s_alloc_mem;
+extern struct st_mysql_plugin spider_i_s_wrapper_protocols;
 #ifdef MARIADB_BASE_VERSION
 extern struct st_maria_plugin spider_i_s_alloc_mem_maria;
+extern struct st_maria_plugin spider_i_s_wrapper_protocols_maria;
 #endif
 
 extern volatile ulonglong spider_mon_table_cache_version;
@@ -932,13 +934,23 @@ static MYSQL_THDVAR_INT(
 SPIDER_THDVAR_OVERRIDE_VALUE_FUNC(int, bulk_update_size)
 
 /*
-  Notes on merge conflicts (remove after merging):
-  10.5: 48faa20db848012e2187a09e05aba832078cb82e
-  10.6: 51ff9eddf7c0aaf1e022fcb3b48ec36835df7785
-  10.9: 06a61b8e453126c2de1649073f247d34e85f9702
-  10.10: 90cd0c156f5bb53fd058d2bbfb83f850ffae6722
-  10.11+: 124eb662700708f3c4b0fb77968f8b854d6bb4aa
-*/
+ -1 :fallback to default
+  0-:buffer size
+ */
+static MYSQL_THDVAR_INT(
+  buffer_size, /* name */
+  PLUGIN_VAR_RQCMDARG, /* opt */
+  "Buffer size", /* comment */
+  NULL, /* check */
+  NULL, /* update */
+  16000, /* def */
+  -1, /* min */
+  2147483647, /* max */
+  0 /* blk */
+);
+
+SPIDER_THDVAR_OVERRIDE_VALUE_FUNC(int, buffer_size)
+
 /*
  -1 :fallback to default
   0 :off
@@ -1563,12 +1575,7 @@ static MYSQL_THDVAR_INT(
   0 /* blk */
 );
 
-double spider_param_hs_ping_interval(
-  THD *thd
-) {
-  DBUG_ENTER("spider_param_hs_ping_interval");
-  DBUG_RETURN(THDVAR(thd, hs_ping_interval));
-}
+SPIDER_THDVAR_VALUE_FUNC(double, hs_ping_interval)
 #endif
 
 /*
@@ -2091,12 +2098,7 @@ static MYSQL_THDVAR_UINT(
   0 /* blk */
 );
 
-uint spider_param_hs_r_conn_recycle_mode(
-  THD *thd
-) {
-  DBUG_ENTER("spider_param_hs_r_conn_recycle_mode");
-  DBUG_RETURN(THDVAR(thd, hs_r_conn_recycle_mode));
-}
+SPIDER_THDVAR_VALUE_FUNC(uint, hs_r_conn_recycle_mode)
 
 /*
   0: weak
@@ -2114,12 +2116,7 @@ static MYSQL_THDVAR_UINT(
   0 /* blk */
 );
 
-uint spider_param_hs_r_conn_recycle_strict(
-  THD *thd
-) {
-  DBUG_ENTER("spider_param_hs_r_conn_recycle_strict");
-  DBUG_RETURN(THDVAR(thd, hs_r_conn_recycle_strict));
-}
+SPIDER_THDVAR_VALUE_FUNC(uint, hs_r_conn_recycle_strict)
 
 /*
   0: no recycle
@@ -2138,12 +2135,7 @@ static MYSQL_THDVAR_UINT(
   0 /* blk */
 );
 
-uint spider_param_hs_w_conn_recycle_mode(
-  THD *thd
-) {
-  DBUG_ENTER("spider_param_hs_w_conn_recycle_mode");
-  DBUG_RETURN(THDVAR(thd, hs_w_conn_recycle_mode));
-}
+SPIDER_THDVAR_VALUE_FUNC(uint, hs_w_conn_recycle_mode)
 
 /*
   0: weak
@@ -2161,12 +2153,7 @@ static MYSQL_THDVAR_UINT(
   0 /* blk */
 );
 
-uint spider_param_hs_w_conn_recycle_strict(
-  THD *thd
-) {
-  DBUG_ENTER("spider_param_hs_w_conn_recycle_strict");
-  DBUG_RETURN(THDVAR(thd, hs_w_conn_recycle_strict));
-}
+SPIDER_THDVAR_VALUE_FUNC(uint, hs_w_conn_recycle_strict)
 
 /*
  -1 :fallback to default
@@ -2823,6 +2810,41 @@ static MYSQL_THDVAR_BOOL(
 
 SPIDER_THDVAR_VALUE_FUNC(bool, sync_sql_mode)
 
+/*
+ -1 : fallback to default
+  0 : do not strict
+  1 : do strict
+ */
+static MYSQL_THDVAR_INT(
+  strict_group_by, /* name */
+  PLUGIN_VAR_RQCMDARG, /* opt */
+  "Use columns in select clause strictly for group by clause",
+  NULL, /* check */
+  NULL, /* update */
+  1, /* def */
+  -1, /* min */
+  1, /* max */
+  0 /* blk */
+);
+
+SPIDER_THDVAR_OVERRIDE_VALUE_FUNC(int, strict_group_by)
+
+/*
+ -1 : use table parameter
+  0 : do not strict
+  1 : do strict
+ */
+static MYSQL_THDVAR_BOOL(
+  direct_aggregate, /* name */
+  PLUGIN_VAR_RQCMDARG, /* opt */
+  "Whether to enable direct aggregate",
+  NULL, /* check */
+  NULL, /* update */
+  TRUE /* def */
+);
+
+SPIDER_THDVAR_VALUE_FUNC(bool, direct_aggregate)
+
 static struct st_mysql_storage_engine spider_storage_engine =
 { MYSQL_HANDLERTON_INTERFACE_VERSION };
 
@@ -2861,6 +2883,7 @@ static struct st_mysql_sys_var* spider_system_variables[] = {
   MYSQL_SYSVAR(bulk_size),
   MYSQL_SYSVAR(bulk_update_mode),
   MYSQL_SYSVAR(bulk_update_size),
+  MYSQL_SYSVAR(buffer_size),
   MYSQL_SYSVAR(internal_optimize),
   MYSQL_SYSVAR(internal_optimize_local),
   MYSQL_SYSVAR(use_flash_logs),
@@ -2976,6 +2999,8 @@ static struct st_mysql_sys_var* spider_system_variables[] = {
   MYSQL_SYSVAR(remote_wait_timeout),
   MYSQL_SYSVAR(wait_timeout),
   MYSQL_SYSVAR(sync_sql_mode),
+  MYSQL_SYSVAR(strict_group_by),
+  MYSQL_SYSVAR(direct_aggregate),
   NULL
 };
 
@@ -2997,7 +3022,8 @@ mysql_declare_plugin(spider)
   0,
 #endif
 },
-spider_i_s_alloc_mem
+spider_i_s_alloc_mem,
+spider_i_s_wrapper_protocols
 mysql_declare_plugin_end;
 
 #ifdef MARIADB_BASE_VERSION
@@ -3017,6 +3043,7 @@ maria_declare_plugin(spider)
   SPIDER_DETAIL_VERSION,
   MariaDB_PLUGIN_MATURITY_STABLE
 },
-spider_i_s_alloc_mem_maria
+spider_i_s_alloc_mem_maria,
+spider_i_s_wrapper_protocols_maria
 maria_declare_plugin_end;
 #endif
