@@ -37,15 +37,15 @@
 #define SPIDER_SYS_XA_COL_CNT 5
 #define SPIDER_SYS_XA_PK_COL_CNT 3
 #define SPIDER_SYS_XA_IDX1_COL_CNT 1
-#define SPIDER_SYS_XA_MEMBER_COL_CNT 18
+#define SPIDER_SYS_XA_MEMBER_COL_CNT 19
 #define SPIDER_SYS_XA_MEMBER_PK_COL_CNT 6
-#define SPIDER_SYS_TABLES_COL_CNT 25
+#define SPIDER_SYS_TABLES_COL_CNT 26
 #define SPIDER_SYS_TABLES_PK_COL_CNT 3
 #define SPIDER_SYS_TABLES_IDX1_COL_CNT 1
 #define SPIDER_SYS_TABLES_UIDX1_COL_CNT 3
-#define SPIDER_SYS_LINK_MON_TABLE_COL_CNT 19
+#define SPIDER_SYS_LINK_MON_TABLE_COL_CNT 20
 #define SPIDER_SYS_LINK_FAILED_TABLE_COL_CNT 4
-#define SPIDER_SYS_XA_FAILED_TABLE_COL_CNT 21
+#define SPIDER_SYS_XA_FAILED_TABLE_COL_CNT 22
 #define SPIDER_SYS_POS_FOR_RECOVERY_TABLE_COL_CNT 7
 #define SPIDER_SYS_TABLE_STS_COL_CNT 11
 #define SPIDER_SYS_TABLE_STS_PK_COL_CNT 2
@@ -56,7 +56,6 @@
 #define SPIDER_SYS_LINK_MON_TABLE_TABLE_NAME_SIZE 64
 #define SPIDER_SYS_LINK_MON_TABLE_LINK_ID_SIZE 64
 
-/* For insertion into `spider_mon_table_cache'. */
 class SPIDER_MON_KEY: public SPIDER_SORT
 {
 public:
@@ -73,7 +72,7 @@ TABLE *spider_open_sys_table(
   const char *table_name,
   int table_name_length,
   bool write,
-  Open_tables_backup *open_tables_backup,
+  SPIDER_Open_tables_backup *open_tables_backup,
   bool need_lock,
   int *error_num
 );
@@ -81,26 +80,29 @@ TABLE *spider_open_sys_table(
 void spider_close_sys_table(
   THD *thd,
   TABLE *table,
-  Open_tables_backup *open_tables_backup,
+  SPIDER_Open_tables_backup *open_tables_backup,
   bool need_lock
 );
 
-bool spider_sys_open_tables(
+#if MYSQL_VERSION_ID < 50500
+#else
+bool spider_sys_open_and_lock_tables(
   THD *thd,
   TABLE_LIST **tables,
-  Open_tables_backup *open_tables_backup
+  SPIDER_Open_tables_backup *open_tables_backup
 );
 
 TABLE *spider_sys_open_table(
   THD *thd,
   TABLE_LIST *tables,
-  Open_tables_backup *open_tables_backup
+  SPIDER_Open_tables_backup *open_tables_backup
 );
 
 void spider_sys_close_table(
   THD *thd,
-  Open_tables_backup *open_tables_backup
+  SPIDER_Open_tables_backup *open_tables_backup
 );
+#endif
 
 int spider_sys_index_init(
   TABLE *table,
@@ -237,11 +239,6 @@ void spider_store_tables_link_status(
   long link_status
 );
 
-void spider_store_link_chk_server_id(
-  TABLE *table,
-  uint32 server_id
-);
-
 void spider_store_binlog_pos_failed_link_idx(
   TABLE *table,
   int failed_link_idx
@@ -407,6 +404,7 @@ int spider_get_sys_tables(
 int spider_get_sys_tables_connect_info(
   TABLE *table,
   SPIDER_SHARE *share,
+  int link_idx,
   MEM_ROOT *mem_root
 );
 
@@ -483,6 +481,7 @@ int spider_get_sys_link_mon_server_id(
 int spider_get_sys_link_mon_connect_info(
   TABLE *table,
   SPIDER_SHARE *share,
+  int link_idx,
   MEM_ROOT *mem_root
 );
 
